/* This file is automatically generated by geninit.sh. DO NOT EDIT! */
/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 2002,2005,2007  Free Software Foundation, Inc.
 *
 *  GRUB is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  GRUB is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GRUB.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <grub/emu/misc.h>

extern void grub_adler32_init (void);
extern void grub_adler32_fini (void);
extern void grub_affs_init (void);
extern void grub_affs_fini (void);
extern void grub_afs_init (void);
extern void grub_afs_fini (void);
extern void grub_bfs_init (void);
extern void grub_bfs_fini (void);
extern void grub_blocklist_init (void);
extern void grub_blocklist_fini (void);
extern void grub_btrfs_init (void);
extern void grub_btrfs_fini (void);
extern void grub_cbfs_init (void);
extern void grub_cbfs_fini (void);
extern void grub_cpio_init (void);
extern void grub_cpio_fini (void);
extern void grub_cpio_be_init (void);
extern void grub_cpio_be_fini (void);
extern void grub_crc64_init (void);
extern void grub_crc64_fini (void);
extern void grub_cryptodisk_init (void);
extern void grub_cryptodisk_fini (void);
extern void grub_disk_init (void);
extern void grub_disk_fini (void);
extern void grub_diskfilter_init (void);
extern void grub_diskfilter_fini (void);
extern void grub_dm_nv_init (void);
extern void grub_dm_nv_fini (void);
extern void grub_exfat_init (void);
extern void grub_exfat_fini (void);
extern void grub_ext2_init (void);
extern void grub_ext2_fini (void);
extern void grub_f2fs_init (void);
extern void grub_f2fs_fini (void);
extern void grub_fat_init (void);
extern void grub_fat_fini (void);
extern void grub_geli_init (void);
extern void grub_geli_fini (void);
extern void grub_gzio_init (void);
extern void grub_gzio_fini (void);
extern void grub_hfs_init (void);
extern void grub_hfs_fini (void);
extern void grub_hfsplus_init (void);
extern void grub_hfsplus_fini (void);
extern void grub_hfspluscomp_init (void);
extern void grub_hfspluscomp_fini (void);
extern void grub_host_init (void);
extern void grub_host_fini (void);
extern void grub_hostfs_init (void);
extern void grub_hostfs_fini (void);
extern void grub_iso9660_init (void);
extern void grub_iso9660_fini (void);
extern void grub_jfs_init (void);
extern void grub_jfs_fini (void);
extern void grub_ldm_init (void);
extern void grub_ldm_fini (void);
extern void grub_loopback_init (void);
extern void grub_loopback_fini (void);
extern void grub_ls_init (void);
extern void grub_ls_fini (void);
extern void grub_luks_init (void);
extern void grub_luks_fini (void);
extern void grub_luks2_init (void);
extern void grub_luks2_fini (void);
extern void grub_lvm_init (void);
extern void grub_lvm_fini (void);
extern void grub_lzopio_init (void);
extern void grub_lzopio_fini (void);
extern void grub_macbless_init (void);
extern void grub_macbless_fini (void);
extern void grub_mdraid09_init (void);
extern void grub_mdraid09_fini (void);
extern void grub_mdraid09_be_init (void);
extern void grub_mdraid09_be_fini (void);
extern void grub_mdraid1x_init (void);
extern void grub_mdraid1x_fini (void);
extern void grub_minix_init (void);
extern void grub_minix_fini (void);
extern void grub_minix2_init (void);
extern void grub_minix2_fini (void);
extern void grub_minix2_be_init (void);
extern void grub_minix2_be_fini (void);
extern void grub_minix3_init (void);
extern void grub_minix3_fini (void);
extern void grub_minix3_be_init (void);
extern void grub_minix3_be_fini (void);
extern void grub_minix_be_init (void);
extern void grub_minix_be_fini (void);
extern void grub_newc_init (void);
extern void grub_newc_fini (void);
extern void grub_nilfs2_init (void);
extern void grub_nilfs2_fini (void);
extern void grub_ntfs_init (void);
extern void grub_ntfs_fini (void);
extern void grub_ntfscomp_init (void);
extern void grub_ntfscomp_fini (void);
extern void grub_odc_init (void);
extern void grub_odc_fini (void);
extern void grub_part_acorn_init (void);
extern void grub_part_acorn_fini (void);
extern void grub_part_amiga_init (void);
extern void grub_part_amiga_fini (void);
extern void grub_part_apple_init (void);
extern void grub_part_apple_fini (void);
extern void grub_part_bsd_init (void);
extern void grub_part_bsd_fini (void);
extern void grub_part_dfly_init (void);
extern void grub_part_dfly_fini (void);
extern void grub_part_dvh_init (void);
extern void grub_part_dvh_fini (void);
extern void grub_part_gpt_init (void);
extern void grub_part_gpt_fini (void);
extern void grub_part_msdos_init (void);
extern void grub_part_msdos_fini (void);
extern void grub_part_plan_init (void);
extern void grub_part_plan_fini (void);
extern void grub_part_sun_init (void);
extern void grub_part_sun_fini (void);
extern void grub_part_sunpc_init (void);
extern void grub_part_sunpc_fini (void);
extern void grub_procfs_init (void);
extern void grub_procfs_fini (void);
extern void grub_raid5rec_init (void);
extern void grub_raid5rec_fini (void);
extern void grub_raid6rec_init (void);
extern void grub_raid6rec_fini (void);
extern void grub_reiserfs_init (void);
extern void grub_reiserfs_fini (void);
extern void grub_romfs_init (void);
extern void grub_romfs_fini (void);
extern void grub_sfs_init (void);
extern void grub_sfs_fini (void);
extern void grub_squash4_init (void);
extern void grub_squash4_fini (void);
extern void grub_tar_init (void);
extern void grub_tar_fini (void);
extern void grub_testload_init (void);
extern void grub_testload_fini (void);
extern void grub_udf_init (void);
extern void grub_udf_fini (void);
extern void grub_ufs1_init (void);
extern void grub_ufs1_fini (void);
extern void grub_ufs1_be_init (void);
extern void grub_ufs1_be_fini (void);
extern void grub_ufs2_init (void);
extern void grub_ufs2_fini (void);
extern void grub_xfs_init (void);
extern void grub_xfs_fini (void);
extern void grub_xnu_uuid_init (void);
extern void grub_xnu_uuid_fini (void);
extern void grub_xzio_init (void);
extern void grub_xzio_fini (void);
extern void grub_zfs_init (void);
extern void grub_zfs_fini (void);
extern void grub_zfscrypt_init (void);
extern void grub_zfscrypt_fini (void);
extern void grub_zfsinfo_init (void);
extern void grub_zfsinfo_fini (void);
void
grub_init_all (void)
{
grub_adler32_init ();
grub_affs_init ();
grub_afs_init ();
grub_bfs_init ();
grub_blocklist_init ();
grub_btrfs_init ();
grub_cbfs_init ();
grub_cpio_init ();
grub_cpio_be_init ();
grub_crc64_init ();
grub_cryptodisk_init ();
grub_disk_init ();
grub_diskfilter_init ();
grub_dm_nv_init ();
grub_exfat_init ();
grub_ext2_init ();
grub_f2fs_init ();
grub_fat_init ();
grub_geli_init ();
grub_gzio_init ();
grub_hfs_init ();
grub_hfsplus_init ();
grub_hfspluscomp_init ();
grub_host_init ();
grub_hostfs_init ();
grub_iso9660_init ();
grub_jfs_init ();
grub_ldm_init ();
grub_loopback_init ();
grub_ls_init ();
grub_luks_init ();
grub_luks2_init ();
grub_lvm_init ();
grub_lzopio_init ();
grub_macbless_init ();
grub_mdraid09_init ();
grub_mdraid09_be_init ();
grub_mdraid1x_init ();
grub_minix_init ();
grub_minix2_init ();
grub_minix2_be_init ();
grub_minix3_init ();
grub_minix3_be_init ();
grub_minix_be_init ();
grub_newc_init ();
grub_nilfs2_init ();
grub_ntfs_init ();
grub_ntfscomp_init ();
grub_odc_init ();
grub_part_acorn_init ();
grub_part_amiga_init ();
grub_part_apple_init ();
grub_part_bsd_init ();
grub_part_dfly_init ();
grub_part_dvh_init ();
grub_part_gpt_init ();
grub_part_msdos_init ();
grub_part_plan_init ();
grub_part_sun_init ();
grub_part_sunpc_init ();
grub_procfs_init ();
grub_raid5rec_init ();
grub_raid6rec_init ();
grub_reiserfs_init ();
grub_romfs_init ();
grub_sfs_init ();
grub_squash4_init ();
grub_tar_init ();
grub_testload_init ();
grub_udf_init ();
grub_ufs1_init ();
grub_ufs1_be_init ();
grub_ufs2_init ();
grub_xfs_init ();
grub_xnu_uuid_init ();
grub_xzio_init ();
grub_zfs_init ();
grub_zfscrypt_init ();
grub_zfsinfo_init ();
}
void
grub_fini_all (void)
{
grub_adler32_fini ();
grub_affs_fini ();
grub_afs_fini ();
grub_bfs_fini ();
grub_blocklist_fini ();
grub_btrfs_fini ();
grub_cbfs_fini ();
grub_cpio_fini ();
grub_cpio_be_fini ();
grub_crc64_fini ();
grub_cryptodisk_fini ();
grub_disk_fini ();
grub_diskfilter_fini ();
grub_dm_nv_fini ();
grub_exfat_fini ();
grub_ext2_fini ();
grub_f2fs_fini ();
grub_fat_fini ();
grub_geli_fini ();
grub_gzio_fini ();
grub_hfs_fini ();
grub_hfsplus_fini ();
grub_hfspluscomp_fini ();
grub_host_fini ();
grub_hostfs_fini ();
grub_iso9660_fini ();
grub_jfs_fini ();
grub_ldm_fini ();
grub_loopback_fini ();
grub_ls_fini ();
grub_luks_fini ();
grub_luks2_fini ();
grub_lvm_fini ();
grub_lzopio_fini ();
grub_macbless_fini ();
grub_mdraid09_fini ();
grub_mdraid09_be_fini ();
grub_mdraid1x_fini ();
grub_minix_fini ();
grub_minix2_fini ();
grub_minix2_be_fini ();
grub_minix3_fini ();
grub_minix3_be_fini ();
grub_minix_be_fini ();
grub_newc_fini ();
grub_nilfs2_fini ();
grub_ntfs_fini ();
grub_ntfscomp_fini ();
grub_odc_fini ();
grub_part_acorn_fini ();
grub_part_amiga_fini ();
grub_part_apple_fini ();
grub_part_bsd_fini ();
grub_part_dfly_fini ();
grub_part_dvh_fini ();
grub_part_gpt_fini ();
grub_part_msdos_fini ();
grub_part_plan_fini ();
grub_part_sun_fini ();
grub_part_sunpc_fini ();
grub_procfs_fini ();
grub_raid5rec_fini ();
grub_raid6rec_fini ();
grub_reiserfs_fini ();
grub_romfs_fini ();
grub_sfs_fini ();
grub_squash4_fini ();
grub_tar_fini ();
grub_testload_fini ();
grub_udf_fini ();
grub_ufs1_fini ();
grub_ufs1_be_fini ();
grub_ufs2_fini ();
grub_xfs_fini ();
grub_xnu_uuid_fini ();
grub_xzio_fini ();
grub_zfs_fini ();
grub_zfscrypt_fini ();
grub_zfsinfo_fini ();
}
